
// Tab Preference Observer
//-----------------------------------------------
tabutils._tabPrefObserver = {
	init: function() {
		window.addEventListener("unload", this, false);
		this.register();

		//Close buttons
		TU_hookCode("gBrowser.mTabContainer.adjustTabstrip",
			[/let tab.*/, 
				function() {
					let tab;
					Array.some(this.childNodes, 
						function(aTab) {
							return !aTab.pinned && !aTab.hasAttribute("faviconized") 
								&& aTab.boxObject.width > 0 && (tab = aTab);
						}
					);
				}
			],
			["this.mCloseButtons", "($& & 0x0f)"],
			["this.mCloseButtons != 3", "(this.mCloseButtons & 0x0f) != 3 && !(this.mCloseButtons & 0x20)"],
			["this._closeWindowWithLastTab", "false", "g"],
			["}", 
				function() {
					this.setAttribute("closeButtonOnPointedTab", (this.mCloseButtons & 0x0f) == 1 || !!(this.mCloseButtons & 0x10));
				}
			]
		);

		//Tab counter
		TU_hookCode("gBrowser.mTabContainer.adjustTabstrip", "}", 
			function() {
				if (this.mAllTabsPopup) {
					let n = gBrowser.mTabs.length - gBrowser._removingTabs.length;
					let m = gBrowser.allTabs.length;
					this.mAllTabsPopup.parentNode.label = m == n ? n : [m, n].join("/");
				}
			}
		);

		//Tab animations
		TU_hookCode("gBrowser.removeTab", 'window.getComputedStyle(aTab).maxWidth == "0.1px"', 'aTab.boxObject.width == 0');

		//Shortcuts
		tabutils.addEventListener(window, "keypress", 
			function(event) {
				//TODO: try to find a way not to hard code these events keystrokes!!!
				if ((event.ctrlKey || event.metaKey) && event.shiftKey && (event.charCode == 90 || event.charCode == 122)) {//Ctrl+Shift+Z
					let popup = document.getElementById('undoCloseTabPopup');
					if (popup.state == "open") {
						popup.hidePopup();
						event.preventDefault();
						event.stopPropagation();
					}
				}
			}
			, true
		);

		//Don't allow drag/dblclick on the tab bar to act on the window
		if ("_update" in TabsInTitlebar) // Compat. with Linux
			TU_hookCode("TabsInTitlebar._update", "!this._dragBindingAlive", "$& && TU_getPref('extensions.tabutils.dragBindingAlive', true)");

		gPrefService.getChildList("extensions.tabutils.", {}).sort().concat(
			[
				"browser.link.open_external",
				"browser.tabs.animate", //Bug 649671
				"browser.tabs.tabClipWidth",
				"browser.tabs.tabMaxWidth",
				"browser.tabs.tabMinWidth",
				"browser.tabs.tabMinHeight"
			]
			).forEach(
				function(aPrefName) {
					this.observe(null, "nsPref:changed", aPrefName);
				}
				, this
			);
	},

	register: function() {
		Services.prefs.addObserver("", this, false);
	},

	unregister: function() {
		Services.prefs.removeObserver("", this);
	},

	get cssRules() {
		delete this.cssRules;
		return this.cssRules = {
			current:   { tab:  tabutils.insertRule('.tabbrowser-tab[selected="true"] {}'),
			text: tabutils.insertRule('.tabbrowser-tab[selected="true"] * {}')},
			unread:    { tab:  tabutils.insertRule('.tabbrowser-tab[unread="true"]:not([selected="true"]) {}'),
			text: tabutils.insertRule('.tabbrowser-tab[unread="true"]:not([selected="true"]) * {}')},
			read:      { tab:  tabutils.insertRule('.tabbrowser-tab:not([unread="true"]):not([selected="true"]) {}'),
			text: tabutils.insertRule('.tabbrowser-tab:not([unread="true"]):not([selected="true"]) * {}')},
			unloaded:  { tab:  tabutils.insertRule(':-moz-any(.tabbrowser-tab, .alltabs-item)[pending] {}'),
			text: tabutils.insertRule(':-moz-any(.tabbrowser-tab, .alltabs-item)[pending] * {}')},
			selected:  { tab:  tabutils.insertRule('.tabbrowser-tab[multiselected] {}'),
			text: tabutils.insertRule('.tabbrowser-tab[multiselected] * {}')},
			protected: { tab:  tabutils.insertRule('.tabbrowser-tab[protected] {}'),
			text: tabutils.insertRule('.tabbrowser-tab[protected] * {}')},
			locked:    { tab:  tabutils.insertRule('.tabbrowser-tab[locked] {}'),
			text: tabutils.insertRule('.tabbrowser-tab[locked] * {}')}
		};
	},

	batching: false,
	observe: function(aSubject, aTopic, aData) {
		if (aTopic != "nsPref:changed" || this.batching)
			return;

		switch (aData) {
			case "browser.tabs.animate": this.animate();return;
			case "browser.tabs.tabClipWidth": this.tabClipWidth();return;
			case "browser.tabs.tabMaxWidth": this.tabMaxWidth();return;
			case "browser.tabs.tabMinWidth": this.tabMinWidth();return;
			case "browser.tabs.tabMinHeight": this.tabMinHeight();return;
			case "browser.link.open_external": // Bug 509664 [Fx10]
				if (gPrefService.prefHasUserValue(aData)) {
					TU_setPref("browser.link.open_newwindow.override.external", TU_getPref(aData));
					gPrefService.clearUserPref(aData);
				}
				return;
		}

		if (!aData.startsWith("extensions.tabutils."))
			return;

		let name = aData.slice(20).replace(".", "_", "g");
		if (name in this) {
			this[name]();
			return;
		}

		//Tab stack coloring
		if (/^extensions.tabutils.colorStack.([0-9A-Fa-f]+)$/.test(aData)) {
			this.updateGroupColor(RegExp.$1, TU_getPref(aData));
			return;
		}

		//Tab highlighting
		if (/^extensions.tabutils.(?:highlight|styles.)([^.]+)$/.test(aData)) {
			let prefName = RegExp.$1.toLowerCase();
			let style = {};
			try {
				style = JSON.parse(TU_getPref("extensions.tabutils.styles." + prefName));
				if (!TU_getPref("extensions.tabutils.highlight" + prefName[0].toUpperCase() + prefName.slice(1)))
					style.bold = style.italic = style.underline = style.strikethrough = style.outline 
						= style.color = style.bgColor = style.opacity = false;
			}
			catch (e) {}

			if (!(prefName in this.cssRules))
				return;

			let tabStyle = this.cssRules[prefName].tab.style;
			tabStyle.setProperty("font-weight", style.bold ? "bold" : "", "");
			tabStyle.setProperty("font-style", style.italic ? "italic" : "", "");
			tabStyle.setProperty("text-decoration", style.underline ? "underline" : style.strikethrough ? "line-through" : "", "");
			tabStyle.setProperty("outline", style.outline ? "1px solid" : "", "");
			tabStyle.setProperty("outline-offset", style.outline ? "-1px" : "", "");
			tabStyle.setProperty("outline-color", style.outline ? style.outlineColorCode : "", "");
			tabStyle.setProperty("-moz-outline-radius", style.outline ? "4px" : "", "");
			tabStyle.setProperty("color", style.color ? style.colorCode : "", "important");
			tabStyle.setProperty("background-image", style.bgColor ? "-moz-linear-gradient(" 
				+ style.bgColorCode + "," + style.bgColorCode + ")" : "", "important");
			tabStyle.setProperty("opacity", style.opacity ? style.opacityCode : "", "");

			let textStyle = this.cssRules[prefName].text.style;
			textStyle.setProperty("color", style.color ? style.colorCode : "", "important");
			textStyle.setProperty("background-image", style.bgColor ? "-moz-linear-gradient(" 
				+ style.bgColorCode + "," + style.bgColorCode + ")" : "", "important");
			return;
		}

		//Custom context menuitems
		if (/^extensions.tabutils.menu.([^.]+)$/.test(aData)) {
			let item = document.getElementById(RegExp.$1);
			if (item)
				item.collapsed = !TU_getPref(aData);
			return;
		}

		if (/^extensions.tabutils.menu.([^.]+).([^.]+)$/.test(aData)) {
			let item = document.getElementById(RegExp.$1);
			if (!item) {
				item = document.createElement("menuitem");
				item.id = RegExp.$1;
				item.collapsed = !TU_getPref("extensions.tabutils.menu." + RegExp.$1);

				if (item.id.toLowerCase().indexOf("alltabs") > -1 && gBrowser.mTabContainer.mAllTabsPopup)
					gBrowser.mTabContainer.mAllTabsPopup.insertBefore(item, document.getElementById("sep_closeAllTabs"));
				else
					gBrowser.tabContextMenu.insertBefore(item, document.getElementById("sep_closeTab"));
			}
			this.setAttribute(item, RegExp.$2, TU_getPref(aData));
			return;
		}

		//Custom shortcut keys
		if (/^extensions.tabutils.shortcut.([^.]+)$/.test(aData)) {
			let key = document.getElementById(RegExp.$1);
			if (key)
				key.setAttribute("disabled", !TU_getPref(aData));
			return;
		}

		if (/^extensions.tabutils.shortcut.([^.]+).([^.]+)$/.test(aData)) {
			let key = document.getElementById(RegExp.$1);
			if (!key) {
				key = document.getElementById("tuKeyset").appendChild(document.createElement("key"));
				key.id = RegExp.$1;
				key.setAttribute("disabled", !TU_getPref("extensions.tabutils.shortcut." + RegExp.$1));
			}
			this.setAttribute(key, RegExp.$2, TU_getPref(aData));
			return;
		}

		//Custom toolbar buttons
		if (/^extensions.tabutils.button.([^.]+)$/.test(aData)) {
			let button = document.getElementById(RegExp.$1);
			if (button)
				button.collapsed = !TU_getPref(aData);
			return;
		}

		if (/^extensions.tabutils.button.(newtab-button|alltabs-button|tabs-closebutton).([^.]+)$/.test(aData)) {
			[
				gBrowser.mTabContainer.mTabstrip.querySelector(".tabs-" + RegExp.$1),
				document.getAnonymousElementByAttribute(gBrowser.mTabContainer, "anonid", RegExp.$1),
				document.getElementById(RegExp.$1 == "newtab-button" ? "new-tab-button" : RegExp.$1)
			].forEach(
				function(button) {
					if (button)
						this.setAttribute(button, RegExp.$2, TU_getPref(aData));
				}
				, this
			);
			return;
		}

		if (/^extensions.tabutils.button.([^.]+).([^.]+)$/.test(aData)) {
			let button = document.getElementById(RegExp.$1) || gNavToolbox.palette.getElementsByAttribute("id", RegExp.$1)[0];
			if (!button) {
				button = document.getElementById("nav-bar").appendChild(document.createElement("toolbarbutton"));
				button.id = RegExp.$1;
				button.image = gBrowser.mFaviconService.defaultFavicon.spec;
				button.className = "toolbarbutton-1 chromeclass-toolbar-additional";
				button.collapsed = !TU_getPref("extensions.tabutils.button." + RegExp.$1);
			}
			this.setAttribute(button, RegExp.$2, TU_getPref(aData));
			return;
		}

		//Inject CSS code
		if (/^extensions.tabutils.css.[^.]+$/.test(aData)) {
			try {
				tabutils.insertRule(TU_getPref(aData));
			}
			catch (e) {}
			return;
		}

		//Inject JS code
		if (/^extensions.tabutils.js.[^.]+$/.test(aData)) {
			try {
				new Function(TU_getPref(aData))();
			}
			catch (e) {}
			return;
		}
	},

	setAttribute: function(aElt, aAttr, aVal) {
		aVal == null ? aElt.removeAttribute(aAttr) : aElt.setAttribute(aAttr, aVal);
		if (aAttr == "insertbefore" || aAttr == "insertafter" || aAttr == "parent") {
			let parentNode = document.getElementById(aElt.getAttribute("parent")) || aElt.parentNode;
			let refNode;
			switch (true) {
				case aElt.getAttribute("insertafter") != "":
					refNode = parentNode.getElementsByAttribute("id", aElt.getAttribute("insertafter"))[0];
					refNode = refNode && refNode.nextSibling;
					break;
				case aElt.getAttribute("insertbefore") != "":
					refNode = parentNode.getElementsByAttribute("id", aElt.getAttribute("insertbefore"))[0];
					break;
			}
			parentNode.insertBefore(aElt, refNode);
		}
		else if (aAttr == "separatorbefore" || aAttr == "separatorafter") {
			let refNode = aAttr == "separatorbefore" ? aElt : aElt.nextSibling;
			if (aElt.localName == "menuitem" || aElt.localName == "menu")
				aElt.parentNode.insertBefore(document.createElement("menuseparator"), refNode);
			else if (aElt.localName == "toolbarbutton" || aElt.localName == "toolbaritem")
				aElt.parentNode.insertBefore(document.createElement("toolbarseparator"), refNode);
		}
	},

	animate: function() {
		gBrowser.mTabContainer.setAttribute("dontanimate", !TU_getPref("browser.tabs.animate"));
	},

	tabClipWidth: function() {
		gBrowser.mTabContainer.mTabClipWidth = TU_getPref("browser.tabs.tabClipWidth");
		gBrowser.mTabContainer.adjustTabstrip();
	},

	tabMaxWidth: function() {
		this._tabWidthRule[0].style.setProperty("max-width", TU_getPref("browser.tabs.tabMaxWidth") + "px", "");
		this._tabWidthRule[1].style.setProperty("width", TU_getPref("browser.tabs.tabMaxWidth") + "px", "");
		gBrowser.mTabContainer.adjustTabstrip();
	},

	tabMinWidth: function() {
		this._tabWidthRule[0].style.setProperty("min-width", TU_getPref("browser.tabs.tabMinWidth") + "px", "");
		gBrowser.mTabContainer.adjustTabstrip();
	},

	tabFitTitle: function() {
		gBrowser.mTabContainer.setAttribute("tabfittitle", TU_getPref("extensions.tabutils.tabFitTitle"));
	},

	tabMinHeight: function() {
		this._tabHeightRule[0].style.setProperty("min-height", TU_getPref("browser.tabs.tabMinHeight") + "px", "important");
		this.tabstripHeight();
	},

	tabstripHeight: function() {
		var tab = gBrowser.mTabContainer.lastChild;
		while (tab && tab.boxObject.height == 0)
			tab = tab.previousSibling;
		if (!tab)
			return;

		var wasSelected = tab.selected;
		var wasPinned = tab.pinned;

		tab.removeAttribute("selected");
		tab.removeAttribute("pinned");
		this._tabHeightRule[1].style.minHeight = "";

		var style = getComputedStyle(tab);
		var height = tab.boxObject.height + parseFloat(style.marginTop) + parseFloat(style.marginBottom);
		this._tabHeightRule[1].style.minHeight = height + "px";

		wasSelected ? tab.setAttribute("selected", true) : tab.removeAttribute("selected");
		wasPinned ? tab.setAttribute("pinned", true) : tab.removeAttribute("pinned");
	},

	get _tabWidthRule() {
		delete this._tabWidthRule;
		return this._tabWidthRule = [
			tabutils.insertRule('.tabbrowser-tab:not([faviconized]) {width: 0; -moz-box-flex: 100;}'),
			tabutils.insertRule('.tabbrowser-arrowscrollbox[orient="vertical"] > scrollbox {}'),
			tabutils.insertRule('#tabbrowser-tabs[orient="vertical"] > .tabbrowser-tab {max-width: none !important; -moz-box-flex: 0;}')
		];
	},

	get _tabHeightRule() {
		delete this._tabHeightRule;
		return this._tabHeightRule = [
			tabutils.insertRule('.tabbrowser-tab, .tabbrowser-arrowscrollbox > .tabs-newtab-button {}'),
			tabutils.insertRule('.tabbrowser-tabs:not([multirow]) .tabbrowser-arrowscrollbox > scrollbox {}')
		];
	},

	closeButtons: function() {
		gBrowser.mTabContainer.mCloseButtons = TU_getPref("extensions.tabutils.closeButtons");
		gBrowser.mTabContainer.adjustTabstrip();
	},

	showTabCounter: function() {
		var allTabsPopup = gBrowser.mTabContainer.mAllTabsPopup;
		if (allTabsPopup)
			allTabsPopup.parentNode.setAttribute("showTabCounter", TU_getPref("extensions.tabutils.showTabCounter"));

		gBrowser.mTabContainer.adjustTabstrip();
		this.tabstripHeight();
	},

	showLeftSpace: function() {
		gBrowser.mTabContainer.setAttribute("showLeftSpace", TU_getPref("extensions.tabutils.showLeftSpace"));
		gBrowser.mTabContainer.adjustTabstrip();
	},

	showRightSpace: function() {
		gBrowser.mTabContainer.setAttribute("showRightSpace", TU_getPref("extensions.tabutils.showRightSpace"));
		gBrowser.mTabContainer.adjustTabstrip();
	},

	statusbarMode: function() {
		switch (TU_getPref("extensions.tabutils.statusbarMode")) {
			case 0: document.getElementById("status-bar").setAttribute("mode", "icons");break;
			case 1: document.getElementById("status-bar").setAttribute("mode", "text");break;
			default: document.getElementById("status-bar").setAttribute("mode", "full");break;
		}
	},

	hideOpenInTab: function() {
		var hideOpenInTab = TU_getPref("extensions.tabutils.hideOpenInTab");
		document.getElementById("statusbar-openintab").collapsed = hideOpenInTab;
	},

	hideLoadInBackground: function() {
		var hideLoadInBackground = TU_getPref("extensions.tabutils.hideLoadInBackground");
		if (hideLoadInBackground)
		TU_setPref("extensions.tabutils.loadAllInBackground", false);
		document.getElementById("statusbar-loadinbackground").collapsed = hideLoadInBackground;
	},

	hideLoadInForeground: function() {
		var hideLoadInForeground = TU_getPref("extensions.tabutils.hideLoadInForeground");
		if (hideLoadInForeground)
		TU_setPref("extensions.tabutils.loadAllInForeground", false);
		document.getElementById("statusbar-loadinforeground").collapsed = hideLoadInForeground;
	},

	openLinkInTab: function() {
		tabutils.gOpenLinkInTab = TU_getPref("extensions.tabutils.openLinkInTab");
		document.getElementById("statusbar-openintab").setAttribute("checked", tabutils.gOpenLinkInTab);
	},

	loadAllInBackground: function() {
		tabutils.gLoadAllInBackground = TU_getPref("extensions.tabutils.loadAllInBackground");
		if (tabutils.gLoadAllInBackground)
			TU_setPref("extensions.tabutils.loadAllInForeground", false);
		document.getElementById("statusbar-loadinbackground").setAttribute("checked", tabutils.gLoadAllInBackground);
	},

	loadAllInForeground: function() {
		tabutils.gLoadAllInForeground = TU_getPref("extensions.tabutils.loadAllInForeground");
		if (tabutils.gLoadAllInForeground)
			TU_setPref("extensions.tabutils.loadAllInBackground", false);
		document.getElementById("statusbar-loadinforeground").setAttribute("checked", tabutils.gLoadAllInForeground);
	},

	loadInNewTab: function() {
		switch (TU_getPref("extensions.tabutils.loadInNewTab")) {
			case 0: TU_setPref("browser.newtab.url", "about:blank");break;
			case 1: TU_setPref("browser.newtab.url", gHomeButton.getHomePage().split("|")[0]);break;
		}
	},

	dragBindingAlive: function() {
		let tabsToolbar = document.getElementById("TabsToolbar");
		if (tabsToolbar._dragBindingAlive != null)
			tabsToolbar._dragBindingAlive = TU_getPref("extensions.tabutils.dragBindingAlive", true);
	},

	pinTab_showPhantom: function() {
		gBrowser.updatePinnedTabsBar();
		gBrowser.mTabContainer.setAttribute("showPhantom", TU_getPref("extensions.tabutils.pinTab.showPhantom"));
		gBrowser.mTabContainer.positionPinnedTabs();
		gBrowser.mTabContainer.adjustTabstrip();
	},

	colorStack: function() {
		gBrowser.mTabContainer.setAttribute("colorStack", TU_getPref("extensions.tabutils.colorStack"));
	},

	toolbarShadowOnTab: "-moz-linear-gradient(bottom, rgba(10%,10%,10%,.4) 1px, transparent 1px)",
	bgTabTexture: "-moz-linear-gradient(transparent, hsla(0,0%,45%,.1) 1px, hsla(0,0%,32%,.2) 80%, hsla(0,0%,0%,.2))",
	bgTabTextureHover: "-moz-linear-gradient(hsla(0,0%,100%,.3) 1px, hsla(0,0%,75%,.2) 80%, hsla(0,0%,60%,.2))",
	selectedTabTexture: "-moz-linear-gradient(rgba(255,255,255,0), rgba(255,255,255,.5) 50%)",

	_tabColoringRules: {},
	updateGroupColor: function(group, color) {
		if (color && !(group in this._tabColoringRules)) {
			let selectorText;
			if (group[0] == "{")
				selectorText = '#main-window .tabbrowser-tab[group="' + group + '"]:not([group-counter="1"])';
			else
				selectorText = '.tabbrowser-tabs[colorStack="true"] > .tabbrowser-tab[group^="{' + group + '"]:not([group-counter="1"])';

			this._tabColoringRules[group] = [
				tabutils.insertRule(selectorText + '{}'),
				tabutils.insertRule(selectorText + ':hover {}'),
				tabutils.insertRule(selectorText + '[selected="true"] {}'),
				tabutils.insertRule('#main-window[tabsontop=false]:not([disablechrome]) ' 
					+ selectorText.replace('#main-window', '#tabbrowser-tabs >') + '[selected="true"]:not(:-moz-lwtheme) {}')
			];
		}

		if (group in this._tabColoringRules) {
			let gradient = '-moz-linear-gradient(' + color + ', -moz-dialog)';
			this._tabColoringRules[group][0].style.backgroundImage = color 
				? [this.toolbarShadowOnTab, this.bgTabTexture, gradient].join() : "";
			this._tabColoringRules[group][1].style.backgroundImage = color 
				? [this.toolbarShadowOnTab, this.bgTabTextureHover, gradient].join() : "";
			this._tabColoringRules[group][2].style.backgroundImage = color 
				? [this.selectedTabTexture, gradient].join() : "";
			this._tabColoringRules[group][3].style.backgroundImage = color 
				? [this.toolbarShadowOnTab, this.selectedTabTexture, gradient].join() : "";
		}
	},

	handleEvent: function(event) {
		switch (event.type) {
			case "load":
				window.removeEventListener("load", this, false);
				this.init();
				break;
			case "unload":
				window.removeEventListener("unload", this, false);
				this.unregister();
				break;
		}
	}
};

// Tags Folder Observer
//-----------------------------------------------
tabutils._tagsFolderObserver = {
	_tags: ["protected", "locked", "faviconized", "pinned", "autoRename", "autoReload", "norestart"],
	_tagIds: [],
	_taggedURIs: [],

	_getIndexForTag: function(aTag) {
		for (let i = 0; i < this._tags.length; i++) {
			if (this._tags[i].toLowerCase() == aTag.toLowerCase())
				return i;
		}
		return -1;
	},

	_updateTaggedURIs: function(aTag, aIndex) {
		if (aIndex == null) {
			aIndex = typeof(aTag) == "string" 
				? this._getIndexForTag(aTag) : this._tagIds.indexOf(aTag);
			if (aIndex == -1)
				return;
			aTag = this._tags[aIndex];
		}

		this._tagIds[aIndex] = -1;
		this._taggedURIs[aIndex] = PlacesUtils.tagging.getURIsForTag(aTag);
		this._tagIds[aIndex] = PlacesUtils.getItemIdForTag(aTag);
	},

	init: function() {
		this._tags.forEach(this._updateTaggedURIs, this);
		PlacesUtils.bookmarks.addObserver(this, false);
	},

	uninit: function() {
		PlacesUtils.bookmarks.removeObserver(this);
	},

	getURIsForTag: function(aTag) {
		let index = this._getIndexForTag(aTag);
		return index > -1 && this._tagIds[index] > -1 ? this._taggedURIs[index] : [];
	},

	getTagsForURI: function(aURI) {
		let tags = [];
		this._tags.forEach(
			function(aTag, aIndex) {
				if (this._tagIds[aIndex] > -1 &&
					this._taggedURIs[aIndex].some(function(bURI) aURI.spec == bURI.spec))
				tags.push(aTag);
			}
			, this
		);
		return tags;
	},

	onItemAdded: function(aItemId, aParentId, aIndex, aItemType, aURI, aTitle/* 6.0 */) {
		if (
			aParentId == PlacesUtils.bookmarks.tagsFolder &&
			aItemType == PlacesUtils.bookmarks.TYPE_FOLDER
		) {
			if (aTitle == null)
				aTitle = PlacesUtils.bookmarks.getItemTitle(aItemId);
			this._updateTaggedURIs(aTitle);
		}
		else if (aItemType == PlacesUtils.bookmarks.TYPE_BOOKMARK) {
			this._updateTaggedURIs(aParentId);
		}
	},

	onItemRemoved: function(aItemId, aParentId, aIndex, aItemType) {
		if ( //TODO: IF condition same as above (onItemAdded)
			aParentId == PlacesUtils.bookmarks.tagsFolder &&
			aItemType == PlacesUtils.bookmarks.TYPE_FOLDER
		) {
			this._updateTaggedURIs(aItemId);
		}
		else if (aItemType == PlacesUtils.bookmarks.TYPE_BOOKMARK) {
			this._updateTaggedURIs(aParentId);
		}
	},

	onItemChanged: function(aItemId, aProperty, aIsAnnotationProperty, aNewValue, aLastModified, aItemType, aParentId/* 6.0 */) {
		if (aParentId == null)
			aParentId = PlacesUtils.bookmarks.getFolderIdForItem(aItemId);

		if (
			aProperty == "title" 
			&& aParentId == PlacesUtils.bookmarks.tagsFolder 
			&& aItemType == PlacesUtils.bookmarks.TYPE_FOLDER
		) {
			this._updateTaggedURIs(aItemId);
			this._updateTaggedURIs(aNewValue);
		}
		else if (aProperty = "uri" && aItemType == PlacesUtils.bookmarks.TYPE_BOOKMARK) {
			this._updateTaggedURIs(aParentId);
		}
	},

	onItemMoved: function(aItemId, aOldParentId, aOldIndex, aNewParentId, aNewIndex, aItemType) {
		if (aItemType == PlacesUtils.bookmarks.TYPE_FOLDER) {
			if (aOldParentId == PlacesUtils.bookmarks.tagsFolder)
				this._updateTaggedURIs(aItemId);
			else if (aNewParentId == PlacesUtils.bookmarks.tagsFolder)
				this._updateTaggedURIs(PlacesUtils.bookmarks.getItemTitle(aItemId));
		}
		else if (aItemType == PlacesUtils.bookmarks.TYPE_BOOKMARK) {
			this._updateTaggedURIs(aOldParentId);
			this._updateTaggedURIs(aNewParentId);
		}
	},

	onBeginUpdateBatch: function() {},
	onEndUpdateBatch: function() {},
	onBeforeItemRemoved: function() {},
	onItemVisited: function() {},
	QueryInterface: XPCOMUtils.generateQI([Ci.nsINavBookmarkObserver])
};
