// Handler code for preferences.xul


//---------------------------------------------------------------------------------------------
// Persistent variables
//---------------------------------------------------------------------------------------------

//List of preferences that need to be updated:
var preferencesToApply = [];

//Constants
var c_defaultPrefFilename = "TUprefs.txt";

//---------------------------------------------------------------------------------------------
// Form Load/unload
//---------------------------------------------------------------------------------------------

//Preferences panel Load handler
function onLoad() {
    if (document.documentElement.getButton("accept").hidden) {
        document.documentElement.getButton("extra1").hidden = true; //hide "apply" button
    }
    else {
        document.documentElement.getButton("extra1").disabled = true; //show "apply" button but keep it disabled
        window.addEventListener("change", onChange, false); //watch for changes in preference values
    }

    //Init "settings" menu on the right
    document.documentElement.getButton("extra2").type = "menu";
    document.documentElement.getButton("extra2").setAttribute("popup", "settingsPopup");

    //clear list of preferences to apply:
    preferencesToApply = [];

    var preferences = document.getElementsByTagName("preference");
    var preferencesElements = document.getElementsByTagName("preferences"); //all preference elements
    var preferencesElement = preferencesElements[preferencesElements.length - 1]; //get last element

    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
    prefs.getChildList("extensions.tabutils.", {}).sort().forEach(function(aPrefName) {
    for (let preference of preferences) {
        if (preference.name == aPrefName)
            return;
    }

    var preference = document.createElement("preference");
    preference.setAttribute("id", aPrefName);
    preference.setAttribute("name", aPrefName);
    preference.setAttribute("type", {32: "string", 64: "int", 128: "bool"}[prefs.getPrefType(aPrefName)]);
        preferencesElement.appendChild(preference);
    });

    var popup = $("tabClickingOptions");
    Array.forEach(document.getElementsByTagName("preference"), function(aPref) {
    if (/^extensions.tabutils.mouse.(\d+).label$/.test(aPref.name) && aPref.value) {
        let item = popup.getElementsByAttribute("value", RegExp.$1)[0];
        if (!item)
            item = popup.appendChild(document.createElement("menuitem"));
        item.setAttribute("value", RegExp.$1);
        item.setAttribute("label", aPref.value);
    }
    });

    ["c.clickTab", "c.clickTabBar", "c.clickNewTabButton"].forEach(function(aId) {
        $(aId).appendChild(popup.cloneNode(true));
    });

    loadAllInBackground();
    loadAllInForeground();
    openTabNext();
    loadInNewTab();
    openNewWindow();
    openNewWindowRestriction();
    singleWindowMode();
    mouseHoverPreview();
    mouseHoverSelect();
    mouseScrollSelect();
    tabBarPosition();
    closeButtons();
    hideLoadInBackground();
    hideLoadInForeground();

    onTabClickingTabSelect();
    onHighlightTabSelect();
    onHighlightStyle("Current");
    onHighlightStyle("Unread");
    onHighlightStyle("Read");
    onHighlightStyle("Unloaded");
    onHighlightStyle("Selected");
    onHighlightStyle("Protected");
    onHighlightStyle("Locked");

    let version = parseFloat(Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULAppInfo).version);
    if (version >= 4.0) {
        $("autoHide").disabled = true;
    }
    else {
        $("context_mergeGroup").disabled = true;
    }
}

function onUnload() {
    //stop watching for changes in preferences window
    window.removeEventListener("change", onChange, false);
}

//---------------------------------------------------------------------------------------------
// Preferences to apply
//---------------------------------------------------------------------------------------------

//add updated preference to list of preferences to apply.
function onChange(event) {
  if (event.target.localName != "preference")
    return;

  var preference = event.target;
  var index = preferencesToApply.indexOf(preference);
  if (index != -1 && preference.value == preference.valueFromPreferences) {
    preferencesToApply.splice(index, 1);
  }
  else if (index == -1 && preference.value != preference.valueFromPreferences) {
    preferencesToApply.push(preference);
  }

  //update "apply/cancel" buttons
  document.documentElement.getButton("extra1").disabled = preferencesToApply.length == 0;
  document.documentElement.getButton("cancel").label = preferencesToApply.length == 0 ?
                                                       document.documentElement.getAttribute("buttonlabelclose") :
                                                       document.documentElement.getAttribute("buttonlabelreset");
  //alt:
  /*
  document.documentElement.getButton("cancel").label = document.documentElement.getAttribute( 
        preferencesToApply.length == 0 ? "buttonlabelclose" : "buttonlabelreset" );
  */
}

//commit changes to preferences
function onApply() {
  var preferences = document.getElementsByTagName("preference");
  for (let preference of preferences) {
    preference.valueFromPreferences = preference.value;
  }

  preferencesToApply = []; //clear update list
  //reset buttons
  document.documentElement.getButton("extra1").disabled = true;
  document.documentElement.getButton("cancel").label = document.documentElement.getAttribute("buttonlabelclose");
}

//restore changes to preferences
function onCancel() {
  if (preferencesToApply.length == 0)
    return true;

  var preferences = document.getElementsByTagName("preference");
  for (let preference of preferences) {
    preference.value = preference.valueFromPreferences;
  }
  return false;
}

//---------------------------------------------------------------------------------------------
// Settings Menu Handlers
//---------------------------------------------------------------------------------------------

//Set default preferences:
function onDefault() {
  var preferences = document.getElementsByTagName("preference");
  for (let preference of preferences) {
    preference.value = preference.defaultValue == null ? undefined : preference.defaultValue;
  }
}

//Apply "recommended" preferences:
function onSuggest() {
  var preferences = document.getElementsByTagName("preference");
  for (let preference of preferences) {
    if (preference.hasAttribute("suggest")) {
      switch (preference.type) {
        case "bool":
          preference.value = (preference.getAttribute("suggest") == "true") != preference.inverted;
          break;
        case "int":
          preference.value = parseInt(preference.getAttribute("suggest"));
          break;
        case "string":
        default:
          preference.value = preference.getAttribute("suggest");
          break;
      }
    }
    else {
      preference.value = preference.defaultValue == null ? undefined : preference.defaultValue;
    }
  }
}

//set preference (this is a handler for Loading each preference
//from the file.)
function pref(aPrefName, aValue) {
  var preferences = document.getElementsByTagName("preference");
  for (let preference of preferences) {
    if (preference.name == aPrefName) {
      preference.value = preference.inverted ? !aValue : aValue;
      return;
    }
  }

  var preferencesElements = document.getElementsByTagName("preferences");
  var preferencesElement = preferencesElements[preferencesElements.length - 1];

  var preference = preferencesElement.appendChild(document.createElement("preference"));
  preference.setAttribute("id", aPrefName);
  preference.setAttribute("name", aPrefName);
  preference.setAttribute("type", {"boolean": "bool", "number": "int"}[typeof aValue] || "string");
  preference.value = aValue;
}



//Save preferences to file
function onExport() {
  var fp = Components.classes['@mozilla.org/filepicker;1'].createInstance(Components.interfaces.nsIFilePicker);
  fp.init(window, "", fp.modeSave);
  fp.defaultString = c_defaultPrefFilename;
  fp.defaultExtension = "txt";
  fp.appendFilters(fp.filterText);
  fp.appendFilters(fp.filterAll);
  if (fp.show() == fp.returnCancel)
    return;

  var str = "";
  var preferences = document.getElementsByTagName("preference");
  for (let preference of preferences) {
    str += 'pref(';
    str += preference.name.quote();
    str += ', ';
    str += preference.type == typeof preference.value ? preference.value.quote() :
           preference.inverted ? !preference.value : preference.value;
    str += ');\n';
  }

  var stream = Components.classes['@mozilla.org/network/file-output-stream;1'].createInstance(Components.interfaces.nsIFileOutputStream);
  stream.init(fp.file, -1, -1, 0);
  stream.write(str, str.length);
  stream.close();
}


//Load preferences from file
function onImport() {
  var fp = Components.classes['@mozilla.org/filepicker;1'].createInstance(Components.interfaces.nsIFilePicker);
  fp.init(window, "", fp.modeOpen);
  fp.defaultString = c_defaultPrefFilename;
  fp.defaultExtension = "txt";
  fp.appendFilters(fp.filterText);
  fp.appendFilters(fp.filterAll);
  if (fp.show() == fp.returnCancel)
    return;

  var ioService = Components.classes['@mozilla.org/network/io-service;1'].getService(Components.interfaces.nsIIOService);
  var loader = Components.classes['@mozilla.org/moz/jssubscript-loader;1'].getService(Components.interfaces.mozIJSSubScriptLoader);
  loader.loadSubScript(ioService.newFileURI(fp.file).spec);
}

//load advanced settings window (using about:config)
function onAdvanced() {
    openDialog('about:config?filter=extensions.tabutils.', '', 'chrome,titlebar,toolbar,centerscreen,dialog=no,resizable').focus();
}

//---------------------------------------------------------------------------------------------
// General Handlers
//---------------------------------------------------------------------------------------------

function openHelp() {
  var win = Components.classes['@mozilla.org/browser/browserglue;1'].getService(Components.interfaces.nsIBrowserGlue).getMostRecentBrowserWindow();
  win.gBrowser.loadOneTab(document.documentElement.getAttribute("helpURI"), null, null, null, false, false);
  win.focus();
}

function loadAllInBackground() {
  setDisabled("b.loadAllInBackgroundOrForeground", $("loadAllInBackground").value || $("loadAllInForeground").value);
  if ($("loadAllInBackground").value)
    $("loadAllInForeground").value = false;
}

function loadAllInForeground() {
  setDisabled("b.loadAllInBackgroundOrForeground", $("loadAllInBackground").value || $("loadAllInForeground").value);
  if ($("loadAllInForeground").value)
    $("loadAllInBackground").value = false;
}

function openTabNext() {
  $("keepOrder").disabled = $("openTabNext").value == 0;
}

function loadInNewTab() {
  setDisabled("b.loadInNewTab", $("loadInNewTab").value != 2);
}

function openNewWindow() {
  setDisabled("b.openNewWindow", $("openNewWindow").value == 2);
}

function openNewWindowRestriction() {
  setDisabled("b.openNewWindowRestriction", $("openNewWindowRestriction").value == 0);
}

function singleWindowMode() {
  var singleWindowMode = $("singleWindowMode").value;
  $("openExternal").value = singleWindowMode && $("openExternal").valueFromPreferences == 2 ? 3 : $("openExternal").valueFromPreferences;
  $("openNewWindow").value = singleWindowMode && $("openNewWindow").valueFromPreferences == 2 ? 3 : $("openNewWindow").valueFromPreferences;
  $("openNewWindowRestriction").value = singleWindowMode ? 0 : $("openNewWindowRestriction").valueFromPreferences;

  setDisabled("b.singleWindowMode", singleWindowMode);
}

function mouseHoverPreview() {
  $("mouseHoverPreviewDelay").disabled = !$("mouseHoverPreview").value;
  $("mouseHoverSelect").disabled = !$("mouseHoverPreview").value;
  $("mouseHoverSelectDelay").disabled = !$("mouseHoverPreview").value || !$("mouseHoverSelect").value;
  $("mouseHoverLoad").disabled = !$("mouseHoverPreview").value;
}

function mouseHoverSelect() {
  $("mouseHoverSelectDelay").disabled = !$("mouseHoverPreview").value || !$("mouseHoverSelect").value;
}

function mouseScrollSelect() {
  var mouseScrollSelect = $("mouseScrollSelect").value;
  $("mouseScrollSelectDir").disabled = !mouseScrollSelect;
  $("mouseScrollSelectWrap").disabled = !mouseScrollSelect;
}

function tabBarPosition() {
  $("showAllTabs").disabled = $("tabBarPosition").value > 1;
}

function closeButtons() {
  $("tabClipWidth").disabled = ($("closeButtons").value & 0x0f) != 1;
}

function hideLoadInBackground() {
  $("loadAllInBackground").disabled = $("hideLoadInBackground").value;
}

function hideLoadInForeground() {
  $("loadAllInForeground").disabled = $("hideLoadInForeground").value;
}

function setDisabled(id, val) {
  val ? $(id).setAttribute("disabled", true) : $(id).removeAttribute("disabled");
}

function $() {
  return document.getElementById.apply(document, arguments);
}

function onTabClickingTabSelect() {
  var target = $("tabClickingTabs");
  var tabs = ["left", "middle", "right", "dbl", "ctrl", "alt", "shift"];

  $("c.clickBookmarks").selectedIndex = target.selectedIndex;
  $("c.clickLinks").selectedIndex = target.selectedIndex;
  $("l.clickTab").selectedIndex = target.selectedIndex ? 1 : 0;

  ["c.clickTab", "c.clickTabBar", "c.clickNewTabButton"].forEach(function(aId) {
    var prefName = tabs[target.selectedIndex] + aId.split(".")[1].replace(/./, function(s) s.toUpperCase());
    var preference = $(prefName);
    var element = $(aId);
    if (preference) {
      element.setAttribute("preference", preference.id);
      preference.setElementValue(element);
    }
    else {
      element.removeAttribute("preference");
      element.value = 0;
    }
    setDisabled(aId.replace("c", "b"), !element.hasAttribute("preference"));
  });

  setDisabled("b.clickBookmarks", ["middle", "right", "ctrl", "shift"].indexOf(tabs[target.selectedIndex]) == -1);
  setDisabled("b.clickLinks", ["middle", "right", "ctrl", "shift"].indexOf(tabs[target.selectedIndex]) == -1);
  setDisabled("b.clickCurrentTab", target.selectedIndex > 0);
}

function onHighlightTabSelect() {
  var target = $("highlight");
  var tabs = ["Current", "Unread", "Read", "Unloaded", "Selected", "Protected", "Locked"];

  $("c.highlightEnabled").setAttribute("preference", "highlight" + tabs[target.selectedIndex]);
  $("c.highlightEnabled").checked = $("highlight" + tabs[target.selectedIndex]).value;

  onHighlightEnabled(tabs[target.selectedIndex]);
}

function onHighlightEnabled(item) {
  var tabs = ["Current", "Unread", "Read", "Unloaded", "Selected", "Protected", "Locked"];
  if (item == tabs[$("highlight").selectedIndex]) {
    setDisabled("b.highlightEnabled", !$("highlight" + item).value);
  }

  onHighlightStyle(item);
}

function onHighlightStyle(item) {
  var style = {};
  try {
    style = JSON.parse($("style" + item).value) || {};
  }
  catch (e) {}

  $("b.highlight" + item).setAttribute("style",
    $("highlight" + item).value ? (style.bold ? "font-weight: bold;" : "")
                                + (style.italic ? "font-style: italic;" : "")
                                + (style.underline ? "text-decoration: underline;" : "")
                                + (style.strikethrough ? "text-decoration: line-through;" : "")
                                + (style.color ? "color:" + style.colorCode + ";" : "")
                                + (style.bgColor ? "-moz-appearance: none;background-color:" + style.bgColorCode + ";" : "")
                                + (style.outline ? "outline: 1px solid;outline-offset: -1px;-moz-outline-radius: 4px;" : "")
                                + (style.outline ? "outline-color:" + style.outlineColorCode + ";" : "")
                                + (style.opacity ? "opacity:" + style.opacityCode + ";" : "")
                                : ""
  );

  var tabs = ["Current", "Unread", "Read", "Unloaded", "Selected", "Protected", "Locked"];
  if (item == tabs[$("highlight").selectedIndex]) {
    $("c.highlightBold").checked = style.bold;
    $("c.highlightItalic").checked = style.italic;
    $("c.highlightUnderline").checked = style.underline;
    $("c.highlightStrikethrough").checked = style.strikethrough;
    $("c.highlightColor").checked = style.color;
    $("c.highlightColorCode").value = style.colorCode;
    $("c.highlightColorPicker").color = style.colorCode;
    $("c.highlightBgColor").checked = style.bgColor;
    $("c.highlightBgColorCode").value = style.bgColorCode;
    $("c.highlightBgColorPicker").color = style.bgColorCode;
    $("c.highlightOutline").checked = style.outline;
    $("c.highlightOutlineColorCode").value = style.outlineColorCode;
    $("c.highlightOutlineColorPicker").color = style.outlineColorCode;
    $("c.highlightOpacity").checked = style.opacity;
    $("c.highlightOpacityCode").value = style.opacityCode * 100;

    onHighlightColor();
    onHighlightBgColor();
    onHighlightOutline();
    onHighlightOpacity();
  }
}

function doHighlightStyle() {
  var style = {
    bold: $("c.highlightBold").checked,
    italic: $("c.highlightItalic").checked,
    underline: $("c.highlightUnderline").checked,
    strikethrough: $("c.highlightStrikethrough").checked,
    color: $("c.highlightColor").checked,
    colorCode: $("c.highlightColorCode").value,
    bgColor: $("c.highlightBgColor").checked,
    bgColorCode: $("c.highlightBgColorCode").value,
    outline: $("c.highlightOutline").checked,
    outlineColorCode: $("c.highlightOutlineColorCode").value,
    opacity: $("c.highlightOpacity").checked,
    opacityCode: $("c.highlightOpacityCode").value / 100
  };

  var tabs = ["Current", "Unread", "Read", "Unloaded", "Selected", "Protected", "Locked"];
  $("style" + tabs[$("highlight").selectedIndex]).value = JSON.stringify(style);
}

function onHighlightColor() {
  setDisabled("b.highlightColor", $("c.highlightColor").disabled || !$("c.highlightColor").checked);
}

function onHighlightBgColor() {
  setDisabled("b.highlightBgColor", $("c.highlightBgColor").disabled || !$("c.highlightBgColor").checked);
}

function onHighlightOutline() {
  setDisabled("b.highlightOutline", $("c.highlightOutline").disabled || !$("c.highlightOutline").checked);
}

function onHighlightOpacity() {
  setDisabled("b.highlightOpacity", $("c.highlightOpacity").disabled || !$("c.highlightOpacity").checked);
}

function onPaneLoad(aPane) { //Bug 687356 - https://bugzilla.mozilla.org/show_bug.cgi?id=687356
  aPane._inited = true;

  let tabbox = aPane.getElementsByTagName('tabbox')[0];
  if (tabbox)
    tabbox.selectedIndex = aPane.getAttribute('lastSelected');
}

function onPaneSelect(aPane) {
  if(!aPane._inited)
    return;

  let tabbox = aPane.getElementsByTagName('tabbox')[0];
  if (tabbox)
    aPane.setAttribute('lastSelected', tabbox.selectedIndex);
}
    
