var tabutils = {

//---------------------------------------------------------------------------------------------
// Initialization and Loading
//---------------------------------------------------------------------------------------------

	init: function() {
		this._tabEventListeners.init();
		this._PlacesUtilsExt();

		this._openUILinkInTab();
		this._openLinkInTab();
		this._singleWindowMode();

		this._tabOpeningOptions();
		this._tabClosingOptions();
		this._tabClickingOptions();

		this._unreadTab();
		this._protectAndLockTab();
		this._faviconizeTab();
		this._pinTab();
		this._phantomTabs();
		this._renameTab();
		this._restartTab();
		this._reloadEvery();
		this._bookmarkTabs();
		this._multiTabHandler();
		this._groupTabs();
		this._previewTab();
		this._miscFeatures();

		!TU_getPref("extensions.tabutils.disableMultirowTabs") && this._multirowTabs();
		!TU_getPref("extensions.tabutils.disableVerticalTabs") && this._verticalTabs();

		window.addEventListener("load", this, false);
		window.addEventListener("unload", this, false);

		if (gBrowser.mTabListeners.length > 0) { // Bug 463384 [Fx5]
			let tabListener = gBrowser.mTabListeners[0];
			gBrowser.browsers[0].webProgress.removeProgressListener(gBrowser.mTabFilters[0]);
			gBrowser.mTabFilters[0].removeProgressListener(gBrowser.mTabListeners[0]);
			gBrowser.mTabListeners[0] = gBrowser.mTabProgressListener(tabListener.mTab, tabListener.mBrowser, tabListener.mBlank);
			gBrowser.mTabFilters[0].addProgressListener(gBrowser.mTabListeners[0], Ci.nsIWebProgress.NOTIFY_ALL);
			gBrowser.browsers[0].webProgress.addProgressListener(gBrowser.mTabFilters[0], Ci.nsIWebProgress.NOTIFY_ALL);
		}

		if (!("privateBrowsingEnabled" in gPrivateBrowsingUI)) { // Bug 799001 [Fx20]
			XPCOMUtils.defineLazyGetter(gPrivateBrowsingUI, "privateBrowsingEnabled", function() {
				return PrivateBrowsingUtils.isWindowPrivate(window);
			});
		}

//    Function.prototype.__defineGetter__("stack", function() {
//      var stack = [];
//      for (let caller = this; caller && stack.length < 15; caller = caller.caller) {
//        stack.push(caller.name);
//      }
//      return stack;
//    });
//
//    TU_hookCode("gBrowser.addTab", "{", "Cu.reportError([arguments.callee.stack, aURI]);");
//    TU_hookCode("gBrowser.moveTabTo", "{", "Cu.reportError([arguments.callee.stack, aTab._tPos, aIndex]);");
//    TU_hookCode("gBrowser._beginRemoveTab", "{", "Cu.reportError([arguments.callee.stack]);");
//    TU_hookCode("gBrowser._endRemoveTab", "{", "Cu.reportError([arguments.callee.stack]);");
//    TU_hookCode("gBrowser._blurTab", "{", "Cu.reportError([arguments.callee.stack]);");
//
//    TU_hookCode("gBrowser.pinTab", "{", "Cu.reportError([arguments.callee.stack]);");
//    TU_hookCode("gBrowser.mTabContainer.adjustTabstrip", "{", "Cu.reportError([arguments.callee.stack]);");
//    TU_hookCode("gBrowser.mTabContainer.positionPinnedTabs", "{", "Cu.reportError([arguments.callee.stack]);");
//    TU_hookCode("gBrowser.mTabContainer.stylePinnedTabs", "{", "Cu.reportError([arguments.callee.stack]);");
	},

	onload: function() {
		this._mainContextMenu();
		this._tabContextMenu();
		this._allTabsPopup();
		this._hideTabBar();
		this._undoCloseTabButton();
		this._tabPrefObserver.init();
		this._tagsFolderObserver.init();

		this._firstRun();
	},

//---------------------------------------------------------------------------------------------
// Tab Attribute Utilities
//---------------------------------------------------------------------------------------------


	getTabValue: function(aTab, aKey) {
		let val = this._ss.getTabValue(aTab, aKey);
		if (!val) {
			let tabData = aTab.linkedBrowser.__SS_data;
			if (tabData)
				val = tabData.attributes[aKey] || tabData[aKey];
		}
		return val == null ? "" : val.toString();
	},

	setTabValue: function() this._ss.setTabValue.apply(this._ss, arguments),
	deleteTabValue: function() this._ss.deleteTabValue.apply(this._ss, arguments),

	setAttribute: function(aTab, aAttr, aVal) {
		aTab.setAttribute(aAttr, aVal);
		this.setTabValue(aTab, aAttr, aVal);
	},

	removeAttribute: function(aTab, aAttr) {
		aTab.removeAttribute(aAttr);
		this.deleteTabValue(aTab, aAttr);
	},

	restoreAttribute: function(aTab, aAttr) {
		let aVal = this.getTabValue(aTab, aAttr);
		if (aVal)
			aTab.setAttribute(aAttr, aVal);
		else
			aTab.removeAttribute(aAttr);
	},

//---------------------------------------------------------------------------------------------
// Something else
//---------------------------------------------------------------------------------------------


	getURIsForTag: function() this._tagsFolderObserver.getURIsForTag.apply(this._tagsFolderObserver, arguments),
	getTagsForURI: function() this._tagsFolderObserver.getTagsForURI.apply(this._tagsFolderObserver, arguments),

	getDomainFromURI: function(aURI, aAllowThirdPartyFixup) {
		try {
			if (typeof aURI == "string")
			aURI = this._URIFixup.createFixupURI(aURI, aAllowThirdPartyFixup);
		}
		catch (e) {}

		try {
			return this._eTLDService.getBaseDomain(aURI);
		}
		catch (e) {}

		try {
			return aURI.host;
		}
		catch (e) {
			return aURI.spec;
		}
	},

	get _styleSheet() {
		for (let i = 0, styleSheet; styleSheet = document.styleSheets[i]; i++) {
			if (styleSheet.href == "chrome://tabutils/skin/tabutils.css") {
				delete this._styleSheet;
				return this._styleSheet = styleSheet;
			}
		}
		return document.styleSheets[0];
	},

	insertRule: function(rule) {
		var ss = this._styleSheet;
		return ss.cssRules[ss.insertRule(rule, ss.cssRules.length)];
	},

	dispatchEvent: function() {
		let event = document.createEvent("Events");
		event.initEvent.apply(event, Array.slice(arguments, 1));
		arguments[0].dispatchEvent(event);
	},

	//GLOBAL VARIABLE: (TODO)
	_eventListeners: [],
	addEventListener: function() {
		document.addEventListener.apply(arguments[0], Array.slice(arguments, 1));
		this._eventListeners.push(arguments);
	},

	onunload: function() {
		this._eventListeners.forEach(function(args) document.removeEventListener.apply(args[0], Array.slice(args, 1)));
		this._tagsFolderObserver.uninit();
	},

	handleEvent: function(event) {
		window.removeEventListener(event.type, this, false);
		switch (event.type) {
			case "DOMContentLoaded": this.init();break;
			case "load": this.onload();break;
			case "unload": this.onunload();break;
		}
	}
};

// End of tabutils definition
//---------------------------------------------------------------------------------------------


//---------------------------------------------------------------------------------------------
// Some Event Listener stuff
//---------------------------------------------------------------------------------------------

window.addEventListener("DOMContentLoaded", tabutils, false);

XPCOMUtils.defineLazyServiceGetter(tabutils, "_ss",
                                   "@mozilla.org/browser/sessionstore;1",
                                   "nsISessionStore");

XPCOMUtils.defineLazyServiceGetter(tabutils, "_URIFixup",
                                   "@mozilla.org/docshell/urifixup;1",
                                   "nsIURIFixup");

XPCOMUtils.defineLazyServiceGetter(tabutils, "_eTLDService",
                                   "@mozilla.org/network/effective-tld-service;1",
                                   "nsIEffectiveTLDService");

tabutils._tabEventListeners = {
	init: function() {
		TU_hookCode("gBrowser.addTab",
			["{", "if (!aURI) aURI = 'about:blank';"],
			[/(?=var evt)/, function() {
				t.arguments = {
					aURI: aURI,
					aReferrerURI: aReferrerURI,
					aRelatedToCurrent: aRelatedToCurrent
				};
			}]
		);

		gBrowser.onTabOpen = function onTabOpen(aTab) {
			var aURI, aReferrerURI, aRelatedToCurrent;
			if (aTab.arguments) {
				aURI = aTab.arguments.aURI;
				aReferrerURI = aTab.arguments.aReferrerURI;
				aRelatedToCurrent = aTab.arguments.aRelatedToCurrent;
			}

			var uri, tags = [];
			try {
				uri = makeURI(aURI);
			}
			catch (e) {
				uri = makeURI("about:blank");
			}

			if (uri.spec != "about:blank")
				tags = tabutils.getTagsForURI(uri, {});
		};

		gBrowser.onLocationChange = function onLocationChange(aTab) {
			var uri = aTab.linkedBrowser.currentURI;
			var tags = tabutils.getTagsForURI(uri, {});
		};

		TU_hookCode("gBrowser.mTabProgressListener", /(?=.*isBlankPageURL.*)/, function() {
			if (!isBlankPageURL(this.mBrowser.currentURI.spec) 
				&& (!this.mBrowser.lastURI || isBlankPageURL(this.mBrowser.lastURI.spec))
				&& !this.mBrowser.__SS_tabStillLoading) // Bug 698565 [Fx11] 
			{
				this.mTabBrowser.onLocationChange(this.mTab);
			}
		});

		TU_hookCode("gBrowser.updateCurrentBrowser", /(?=.*createEvent.*)/
			, (function() {
				if (!oldTab.selected) {
					let event = document.createEvent("Events");
					event.initEvent("TabBlur", true, false);
					oldTab.dispatchEvent(event);
				}
			}).toString().replace(/^.*{|}$/g, "") //TODO: what does this do!?
		);

		gBrowser.onTabMove = function onTabMove(aTab, event) {};
		gBrowser.onTabClose = function onTabClose(aTab) {};
		gBrowser.onTabSelect = function onTabSelect(aTab) {};
		gBrowser.onTabBlur = function onTabBlur(aTab) {};
		gBrowser.onTabPinning = function onTabPinning(aTab) {};
		gBrowser.onTabPinned = function onTabPinned(aTab) {};
		gBrowser.onTabHide = function onTabHide(aTab) {};
		gBrowser.onTabShow = function onTabShow(aTab) {};
		gBrowser.onTabStacked = function onTabStacked(aTab) {};
		gBrowser.onTabUnstacked = function onTabUnstacked(aTab) {};
		gBrowser.onStackCollapsed = function onStackCollapsed(aTab) {};
		gBrowser.onStackExpanded = function onStackExpanded(aTab) {};
		gBrowser.onTabRestoring = function onTabRestoring(aTab) {var ss = tabutils._ss;};
		gBrowser.onTabRestored = function onTabRestored(aTab) {var ss = tabutils._ss;};
		gBrowser.onTabClosing = function onTabClosing(aTab) {var ss = tabutils._ss;};

		[
			"TabOpen", "TabMove", "TabClose", "TabSelect", "TabBlur",
			"TabPinning", "TabPinned", "TabHide", "TabShow",
			"TabStacked", "TabUnstacked", "StackCollapsed", "StackExpanded",
			"SSTabRestoring", "SSTabRestored", "SSTabClosing"
		].forEach(function(type) {
			tabutils.addEventListener(gBrowser.mTabContainer, type, this, false);
		}, this); //TODO: what the heck.... (Interesting way of adding even listeners...)
	},

	handleEvent: function(event) {
		switch (event.type) {
			case "TabOpen": gBrowser.onTabOpen(event.target);break;
			case "TabMove": gBrowser.onTabMove(event.target, event);break;
			case "TabClose": gBrowser.onTabClose(event.target);break;
			case "TabSelect": gBrowser.onTabSelect(event.target);break;
			case "TabBlur": gBrowser.onTabBlur(event.target);break;
			case "TabPinning": gBrowser.onTabPinning(event.target);break;
			case "TabPinned": gBrowser.onTabPinned(event.target);break;
			case "TabHide": gBrowser.onTabHide(event.target);break;
			case "TabShow": gBrowser.onTabShow(event.target);break;
			case "TabStacked": gBrowser.onTabStacked(event.target);break;
			case "TabUnstacked": gBrowser.onTabUnstacked(event.target);break;
			case "StackCollapsed": gBrowser.onStackCollapsed(event.target);break;
			case "StackExpanded": gBrowser.onStackExpanded(event.target);break;
			case "SSTabRestoring": gBrowser.onTabRestoring(event.target);break;
			case "SSTabRestored": gBrowser.onTabRestored(event.target);break;
			case "SSTabClosing": gBrowser.onTabClosing(event.target);break;
		}
	}
};
